use async_channel::{Receiver, Sender};
use leptos::logging::log;
use protocol::{Error, Signal, SignalType};
use wasm_bindgen_futures::spawn_local;
use web_sys::{
    ErrorEvent, MessageEvent, WebSocket, js_sys,
    wasm_bindgen::{JsCast, prelude::Closure},
};

use crate::sleep;

static SIGNALLING_ADDRESS: &str = "http://192.168.1.3:4546/signal";

thread_local! {
    static WEBSOCKET: WebSocket = SignallingChannel::init().unwrap();

    static OFFER_CHANNEL: (Sender<Signal>, Receiver<Signal>) = async_channel::unbounded();
    static ANSWER_CHANNEL: (Sender<Signal>, Receiver<Signal>) = async_channel::unbounded();
    static ICE_CANDIDATE_CHANNEL: (Sender<Signal>, Receiver<Signal>) = async_channel::unbounded();
}

struct SignallingChannel {}

impl SignallingChannel {
    fn is_ready() -> bool {
        if WEBSOCKET.with(|websocket| websocket.ready_state()) == 1 {
            true
        } else {
            false
        }
    }
    fn init() -> Result<WebSocket, Error> {
        let web_socket =
            WebSocket::new(SIGNALLING_ADDRESS).map_err(|_| Error::WebSocketInitialization)?;

        let on_message_callback = Closure::<dyn Fn(_)>::new(move |message_event: MessageEvent| {
            if let Ok(received_data) = message_event.data().dyn_into::<js_sys::JsString>() {
                if let Some(received_data) = received_data.as_string() {
                    if let Ok(received_signal) = serde_json::from_str::<Signal>(&received_data) {
                        let received_signal_type = received_signal.get_signal_type();
                        if received_signal_type == SignalType::ICECandidate {
                            let ice_candidate_channel_sender = ICE_CANDIDATE_CHANNEL
                                .with(|ice_candidate_channel| ice_candidate_channel.0.clone());
                            spawn_local(async move {
                                ice_candidate_channel_sender
                                    .send(received_signal)
                                    .await
                                    .expect("Never")
                            });
                        } else if received_signal_type == SignalType::Offer {
                            let offer_channel_sender =
                                ICE_CANDIDATE_CHANNEL.with(|offer_channel| offer_channel.0.clone());
                            spawn_local(async move {
                                offer_channel_sender
                                    .send(received_signal)
                                    .await
                                    .expect("Never")
                            });
                        } else if received_signal_type == SignalType::Answer {
                            let answer_channel_sender = ICE_CANDIDATE_CHANNEL
                                .with(|answer_channel| answer_channel.0.clone());
                            spawn_local(async move {
                                answer_channel_sender
                                    .send(received_signal)
                                    .await
                                    .expect("Never")
                            });
                        }
                    }
                }
            }
        });
        web_socket.set_onmessage(Some(on_message_callback.as_ref().unchecked_ref()));
        on_message_callback.forget();

        let on_error_callback = Closure::<dyn Fn(_)>::new(move |error_event: ErrorEvent| {
            log!("Error: WebSocket | {:#?}", error_event);
        });
        web_socket.set_onerror(Some(on_error_callback.as_ref().unchecked_ref()));
        on_error_callback.forget();

        Ok(web_socket)
    }

    fn send(signal: &Signal) -> Result<(), Error> {
        let json = serde_json::json!(signal);
        WEBSOCKET.with(|websocket| {
            websocket
                .send_with_str(&json.to_string())
                .map_err(|err_val| Error::WebSocketSend(format!("{:?}", err_val)))
        })
    }

    async fn receive_offer() -> Result<Signal, Error> {
        OFFER_CHANNEL
            .with(|offer_channel| offer_channel.1.clone())
            .recv()
            .await
            .map_err(|err_val| Error::OfferChannelReceive(err_val.to_string()))
    }

    async fn receive_answer() -> Result<Signal, Error> {
        ANSWER_CHANNEL
            .with(|answer_channel| answer_channel.1.clone())
            .recv()
            .await
            .map_err(|err_val| Error::OfferChannelReceive(err_val.to_string()))
    }

    async fn receive_ice_candidate() -> Result<Signal, Error> {
        ICE_CANDIDATE_CHANNEL
            .with(|ice_candidate_channel| ice_candidate_channel.1.clone())
            .recv()
            .await
            .map_err(|err_val| Error::OfferChannelReceive(err_val.to_string()))
    }
}

pub async fn wait_until_communication_is_ready() {
    while !SignallingChannel::is_ready() {
        sleep(100).await;
    }
}

pub fn send_auth(data: &String) -> Result<(), Error> {
    let offer = Signal::new(&SignalType::Auth, data);

    SignallingChannel::send(&offer)
}

pub fn send_offer(data: &String) -> Result<(), Error> {
    let offer = Signal::new(&SignalType::Offer, data);

    SignallingChannel::send(&offer)
}

pub async fn receive_offer() -> Result<Signal, Error> {
    SignallingChannel::receive_offer().await
}

pub fn send_answer(data: &String) -> Result<(), Error> {
    let offer = Signal::new(&SignalType::Answer, data);

    SignallingChannel::send(&offer)
}

pub async fn receive_answer() -> Result<Signal, Error> {
    SignallingChannel::receive_answer().await
}

pub fn send_ice_candidate(data: &String) -> Result<(), Error> {
    let offer = Signal::new(&SignalType::ICECandidate, data);

    SignallingChannel::send(&offer)
}

pub async fn receive_ice_candidate() -> Result<Signal, Error> {
    SignallingChannel::receive_ice_candidate().await
}
