use std::sync::Arc;

use leptos::logging::log;
use protocol::Error;
use wasm_bindgen_futures::{JsFuture, spawn_local};
use web_sys::{
    MediaStream, RtcConfiguration, RtcIceCandidate, RtcIceCandidateInit, RtcIceServer,
    RtcPeerConnection, RtcPeerConnectionIceEvent, RtcPeerConnectionState, RtcSdpType,
    RtcSessionDescriptionInit,
    js_sys::{Array, Reflect},
    wasm_bindgen::{JsCast, JsValue, prelude::Closure},
};

use crate::signal::{
    receive_answer, receive_ice_candidate, receive_offer, send_answer, send_ice_candidate,
    send_offer,
};

pub struct WebRTC {
    peer_connection: RtcPeerConnection,
    audio_stream: Option<MediaStream>,
    video_stream: Option<MediaStream>,
    screen_stream: Option<MediaStream>,
}

impl WebRTC {
    pub fn new(
        audio_stream: Option<MediaStream>,
        video_stream: Option<MediaStream>,
        screen_stream: Option<MediaStream>,
    ) -> Result<Arc<Self>, Error> {
        let ice_server_addresses = vec![JsValue::from("stun:stun.l.google.com:19302")]
            .into_iter()
            .collect::<Array>();
        let ice_server = RtcIceServer::new();
        ice_server.set_urls(&ice_server_addresses);
        let ice_servers = vec![ice_server].into_iter().collect::<Array>();
        let rtc_configuration = RtcConfiguration::new();
        rtc_configuration.set_ice_servers(&ice_servers);

        let peer_connection = RtcPeerConnection::new_with_configuration(&rtc_configuration)
            .map_err(|_| Error::WebRTCInitialization)?;

        let on_ice_candidate = Closure::<dyn Fn(_)>::new(
            move |peer_connection_ice_event: RtcPeerConnectionIceEvent| {
                if let Some(candidate) = peer_connection_ice_event.candidate() {
                    if let Err(err_val) = send_ice_candidate(&candidate.candidate()) {
                        log!("Error: Send ICE Candidate | {}", err_val);
                    }
                }
            },
        );

        peer_connection.set_onicecandidate(Some(on_ice_candidate.as_ref().unchecked_ref()));
        on_ice_candidate.forget();

        let webrtc = Self {
            peer_connection,
            audio_stream,
            video_stream,
            screen_stream,
        };
        let webrtc = Arc::new(webrtc);

        webrtc.add_streams();
        let webrtc_ice_candidate_receiver = webrtc.clone();
        spawn_local(async move { webrtc_ice_candidate_receiver.ice_candidate_receiver().await });

        Ok(webrtc)
    }

    async fn ice_candidate_receiver(&self) {
        while let Ok(received_ice_candidate) = receive_ice_candidate().await {
            let received_ice_candidate =
                RtcIceCandidateInit::new(&received_ice_candidate.get_data());
            if let Ok(received_ice_candidate) = RtcIceCandidate::new(&received_ice_candidate) {
                let add_received_ice_candidate_promise = self
                    .peer_connection
                    .add_ice_candidate_with_opt_rtc_ice_candidate(Some(&received_ice_candidate));
                if let Err(err_val) = JsFuture::from(add_received_ice_candidate_promise)
                    .await
                    .map_err(|_| Error::ICECandidateAdd)
                {
                    log!("Error: Add ICE Candidate | {}", err_val);
                }
            }
        }
    }

    pub fn get_status(&self) -> RtcPeerConnectionState {
        self.peer_connection.connection_state()
    }

    fn add_streams(&self) {
        if let Some(audio_stream) = self.audio_stream.as_ref() {
            self.peer_connection.add_stream(&audio_stream);
        }
        if let Some(video_stream) = self.video_stream.as_ref() {
            self.peer_connection.add_stream(&video_stream);
        }
        if let Some(screen_stream) = self.screen_stream.as_ref() {
            self.peer_connection.add_stream(&screen_stream);
        }
    }

    pub async fn offer(&self) -> Result<(), Error> {
        let offer_promise = self.peer_connection.create_offer();
        match JsFuture::from(offer_promise)
            .await
            .map_err(|_| Error::WebRTCOffer)
        {
            Ok(offer) => {
                let offer_session_description_protocol =
                    Reflect::get(&offer, &JsValue::from_str("sdp"))
                        .map_err(|_| Error::WebRTCSessionDescriptionProtocol)?;
                match offer_session_description_protocol
                    .as_string()
                    .ok_or_else(|| Error::WebRTCOffer)
                {
                    Ok(offer_session_description_protocol) => {
                        let offer = RtcSessionDescriptionInit::new(RtcSdpType::Offer);
                        offer.set_sdp(&offer_session_description_protocol);
                        let set_local_description_promise =
                            self.peer_connection.set_local_description(&offer);

                        JsFuture::from(set_local_description_promise)
                            .await
                            .map_err(|_| Error::WebRTCSetLocalDescription)?;

                        send_offer(&offer_session_description_protocol)?;
                        if let Ok(received_answer) = receive_answer().await {
                            let answer = RtcSessionDescriptionInit::new(RtcSdpType::Answer);
                            answer.set_sdp(&received_answer.get_data());

                            let set_remote_description_promise =
                                self.peer_connection.set_remote_description(&answer);
                            JsFuture::from(set_remote_description_promise)
                                .await
                                .map_err(|_| Error::WebRTCSetRemoteDescription)?;

                            return Ok(());
                        }
                    }
                    Err(err_val) => {
                        log!(
                            "Error: WebRTC Offer | Session Description Protocol String Transformation | {}",
                            err_val
                        );
                    }
                }
            }
            Err(err_val) => {
                log!("Error: WebRTC Offer | Promise | {}", err_val);
            }
        }
        return Err(Error::WebRTCOffer);
    }

    pub async fn answer(&self) -> Result<(), Error> {
        if let Ok(received_offer) = receive_offer().await {
            let offer = RtcSessionDescriptionInit::new(RtcSdpType::Offer);
            offer.set_sdp(&received_offer.get_data());

            let set_remote_description_promise =
                self.peer_connection.set_remote_description(&offer);
            JsFuture::from(set_remote_description_promise)
                .await
                .map_err(|_| Error::WebRTCSetRemoteDescription)?;

            let answer_promise = self.peer_connection.create_answer();
            match JsFuture::from(answer_promise)
                .await
                .map_err(|_| Error::WebRTCAnswer)
            {
                Ok(answer) => {
                    let answer_session_description_protocol =
                        Reflect::get(&answer, &JsValue::from_str("sdp"))
                            .map_err(|_| Error::WebRTCSessionDescriptionProtocol)?;
                    match answer_session_description_protocol
                        .as_string()
                        .ok_or_else(|| Error::WebRTCAnswer)
                    {
                        Ok(answer_session_description_protocol) => {
                            let answer = RtcSessionDescriptionInit::new(RtcSdpType::Answer);
                            answer.set_sdp(&answer_session_description_protocol);

                            let set_local_description_promise =
                                self.peer_connection.set_local_description(&answer);
                            JsFuture::from(set_local_description_promise)
                                .await
                                .map_err(|_| Error::WebRTCSetLocalDescription)?;

                            send_answer(&answer_session_description_protocol)?;

                            return Ok(());
                        }
                        Err(err_val) => {
                            log!(
                                "Error: WebRTC Answer | Session Description Protocol String Transformation | {}",
                                err_val
                            );
                        }
                    }
                }
                Err(err_val) => {
                    log!("Error: WebRTC Answer | Promise | {}", err_val);
                }
            }
        }
        Err(Error::WebRTCAnswer)
    }
}
